use std::cmp;
use std::time::{Duration, Instant};

use bitvec::macros::internal::funty::Fundamental;
use ethereum_types::H256;
use ethnum::U256;
use tracing::log::debug;
use vdf::{InvalidIterations, InvalidProof, VDFParams, WesolowskiVDFParams, VDF};

use crate::h256_to_u256;
use crate::models::BlockExecutionSpec;

// 10 sec = 65000 on modern server CPU (E5-2630 v3 @ 2.40GHz)
// almost the same for Pietrzak & Wesolowski
const VDF_DIFFICULTY_PER_SEC: u64 = 6_500;
const INT_SIZE_BITS: u16 = 2048;

// vdf-0.1.0/src/proof_wesolowski.rs:271:5
// const MAXIMUM_ITERATIONS: u64 = (1u64 << 53) - 1;
const MAXIMUM_ITERATIONS: u64 = 800_000_000; //approximately one day

/// Compute the delay parameter in seconds of a VDF to run
pub fn compute_vdf_difficulty(network_difficulty: u64, hash: H256, quantity: U256) -> u64 {
    let hashed_part = h256_to_u256(hash).as_u64() % MAXIMUM_ITERATIONS;
    let vdf_difficulty = (network_difficulty.max(1).wrapping_mul(hashed_part)) / quantity.as_u64();
    vdf_difficulty
}

const VDF_MIN_INTERVAL: u64 = 1000;
const VDF_MAX_INTERVAL: u64 = 800_000_000;

/// Quantity would be : VDF_MAX_INTERVAL / (2 * VDF_PER_SEC * desired_average_time)

pub fn compute_vdf_difficulty_bounded_option(hash: H256, quantity: U256) -> u64 {
    let hashed_part = h256_to_u256(hash).as_u64() % (VDF_MAX_INTERVAL - VDF_MIN_INTERVAL);
    let vdf_difficulty = (hashed_part + VDF_MIN_INTERVAL) / quantity.as_u64();
    vdf_difficulty
}

pub fn compute_vdf_difficulty_modulo_option(
    hash: H256,
    quantity: U256,
    network_difficulty: u64,
) -> u64 {
    let hashed_part = h256_to_u256(hash).as_u64() % network_difficulty;
    hashed_part / quantity.as_u64()
}

pub fn update_network_difficulty(
    block_spec: &BlockExecutionSpec,
    prev_network_difficulty: U256,
    avg_mining_time_in_ms: u64,
) -> U256 {
    let prev_network_difficulty: u64 = prev_network_difficulty.as_u64();
    let multiplicator: f64 = block_spec.params.desired_average_mining_time_in_ms.as_f64()
        / avg_mining_time_in_ms.as_f64();
    let updated_network_difficulty = (prev_network_difficulty.as_f64() * multiplicator).as_u64();
    U256::from(updated_network_difficulty)
}

/// Is used as a modulus of a uniformly distributed function over [a;b], thus, averaging at (b - a) / 2
pub fn initialize_network_difficulty(block_spec: &BlockExecutionSpec) -> u64 {
    let network_difficulty = ((block_spec.params.desired_average_mining_time_in_ms / 1_000)
        * VDF_DIFFICULTY_PER_SEC)
        * 2;
    network_difficulty
}

/// Launch a VDF and returns its output.
/// Ensure vdf_difficulty is not too high before launching the VDF.
/// `vdf_difficulty` for Wesolowski needs to be between **1u128 << 53** (9_007_199_254_740_992) and **usize::MAX as u64** (18_446_744_073_709_551_615usize)
pub async fn vdf<'a>(
    challenge: Vec<u8>,
    vdf_difficulty: u64,
) -> Result<Vec<u8>, InvalidIterations> {
    // An instance of the VDF.
    // TODO:  Instances can be used arbitrarily many times.
    let wesolowski_vdf = WesolowskiVDFParams(INT_SIZE_BITS).new();
    tokio::spawn(async move {
        let difficulty_check = wesolowski_vdf.check_difficulty(vdf_difficulty);
        if difficulty_check.is_err() {
            return Err(difficulty_check.err().unwrap());
        }
        debug!(
            "Solving wesolowski_vdf, estimated time: {:?} (={:?}min)",
            Duration::from_secs(vdf_difficulty / VDF_DIFFICULTY_PER_SEC),
            vdf_difficulty / VDF_DIFFICULTY_PER_SEC / 60
        );
        let start = Instant::now();
        let result = wesolowski_vdf.solve(&*challenge, vdf_difficulty);
        debug!(
            "Solved wesolowski_vdf in {:?} (estimated: {:?})",
            start.elapsed(),
            Duration::from_secs(vdf_difficulty / VDF_DIFFICULTY_PER_SEC)
        );
        result
    })
    .await
    .unwrap()
}

/// Verify a VDF output
pub fn verify(
    vdf_proof_challenge: &[u8; 32],
    vdf_difficulty: u64,
    vdf_proof_result: &[u8; 516],
) -> Result<(), InvalidProof> {
    let wesolowski_vdf = WesolowskiVDFParams(INT_SIZE_BITS).new();
    wesolowski_vdf.verify(vdf_proof_challenge, vdf_difficulty, vdf_proof_result)
}

type IterationValue = u32;
type HashOutput = [u8];

#[cfg(test)]
mod tests {
    use bitvec::macros::internal::funty::Numeric;
    use hex_literal::hex;
    use statistical::{mean, median, population_standard_deviation};

    use crate::consensus::difficulty;
    use crate::crypto::keccak256;
    use crate::models::{Params, Revision};

    use super::*;

    #[test]
    fn compute_various_vdf_difficulties() {
        let block_spec = BlockExecutionSpec {
            revision: Revision::Frontier,
            active_transitions: Default::default(),
            params: Params {
                chain_id: Default::default(),
                network_id: Default::default(),
                additional_forks: Default::default(),
                expiration_time_limit: 0,
                desired_average_mining_time_in_ms: 10_000, // Used in vdf network difficulty updates
                desired_demurrage_percent_per_day: 0,
                difficulty_adjustment_period: 0,
                min_update_network_difficulty_percent: 1, // Used in vdf network difficulty updates
                max_update_network_difficulty_percent: 5, // Used in vdf network difficulty updates
                transaction_fees_percentage: 0,
                miner_fixed_reward_per_transaction: Default::default(),
            },
            system_contract_changes: Default::default(),
            balance_changes: Default::default(),
            miner_behavior_reward_per_quantity: Default::default(),
            behavior_reward_per_quantity: Default::default(),
            miner_included_behavior_reward_per_quantity: Default::default(),
        };
        let mut classic_network_difficulty: u64 = 1;
        let mut network_difficulty_as_modulus = initialize_network_difficulty(&block_spec);

        let mut classic_average_mining_time = block_spec.params.desired_average_mining_time_in_ms;
        let mut modulus_average_mining_time = block_spec.params.desired_average_mining_time_in_ms;

        let mut classic_mining_difficulty: Vec<f64> = vec![];
        let mut modulus_mining_difficulty: Vec<f64> = vec![];
        let mut bounded_mining_difficulty: Vec<f64> = vec![];

        let quantity: U256 = U256::from(10_000_u64); // Used in testnet injector to reduce interval of hash output to vdf right difficulty
        let quantity_using_modulus = U256::from(1_u64); // Using 1 as base to get to the desired average time for a miner with neutral quantity
        let quantity_using_bounded_computation = U256::from(6150_u64); //to get to a desired average time of 10 seconds
        for i in 1..1_000_000 {
            let hash = keccak256(i.to_ne_bytes());

            let vdf_difficulty_modulus = compute_vdf_difficulty_modulo_option(
                hash.clone(),
                quantity_using_modulus.clone(),
                network_difficulty_as_modulus,
            );
            modulus_average_mining_time = ((modulus_average_mining_time * i)
                + ((vdf_difficulty_modulus / VDF_DIFFICULTY_PER_SEC) * 1_000))
                / (i + 1);

            modulus_mining_difficulty.push(vdf_difficulty_modulus as f64);
            network_difficulty_as_modulus = update_network_difficulty(
                &block_spec,
                U256::from(network_difficulty_as_modulus),
                modulus_average_mining_time,
            )
            .as_u64();

            let classic_vdf_difficulty =
                compute_vdf_difficulty(classic_network_difficulty, hash.clone(), quantity.clone());
            classic_average_mining_time = ((classic_average_mining_time * i)
                + (classic_vdf_difficulty / VDF_DIFFICULTY_PER_SEC) * 1_000)
                / (i + 1);
            classic_mining_difficulty.push(classic_vdf_difficulty as f64);
            /*classic_network_difficulty = difficulty::update_network_difficulty(
                &block_spec,
                U256::from(classic_network_difficulty),
                classic_average_mining_time + 1, //otherwise may fail, divide by zero
            )
            .as_u64();*/

            bounded_mining_difficulty.push(compute_vdf_difficulty_bounded_option(
                hash,
                quantity_using_bounded_computation.clone(),
            ) as f64);
        }
        // Calculer moyenne, ecart type, mediane, 1/3 quartiles
        let classic_standard_deviation =
            population_standard_deviation(&classic_mining_difficulty, None)
                / VDF_DIFFICULTY_PER_SEC as f64;
        let classic_mean = mean(&classic_mining_difficulty) / VDF_DIFFICULTY_PER_SEC as f64;
        let classic_median = median(&classic_mining_difficulty) / VDF_DIFFICULTY_PER_SEC as f64;

        println!("Valeurs pour le calcul original de difficulté :");
        println!("  - Moyenne = {:.2}s", classic_mean);
        println!("  - Ecart type = {:.2}s", classic_standard_deviation);
        println!("  - Médiane = {:.2}s", classic_median);

        let modulus_standard_deviation =
            population_standard_deviation(&modulus_mining_difficulty, None)
                / VDF_DIFFICULTY_PER_SEC as f64;
        let modulus_mean = mean(&modulus_mining_difficulty) / VDF_DIFFICULTY_PER_SEC as f64;
        let modulus_median = median(&modulus_mining_difficulty) / VDF_DIFFICULTY_PER_SEC as f64;
        println!("Valeurs pour le calcul avec la difficulté de réseau en modulo :");
        println!("  - Moyenne = {:.2}s", modulus_mean);
        println!("  - Ecart type = {:.2}s", modulus_standard_deviation);
        println!("  - Médiane = {:.2}s", modulus_median);

        let bounded_standard_deviation =
            population_standard_deviation(&bounded_mining_difficulty, None)
                / VDF_DIFFICULTY_PER_SEC as f64;
        let bounded_mean = mean(&bounded_mining_difficulty) / VDF_DIFFICULTY_PER_SEC as f64;
        let bounded_median = median(&bounded_mining_difficulty) / VDF_DIFFICULTY_PER_SEC as f64;
        println!("Valeurs pour le calcul dans un intervalle défini sans difficulté de réseau :");
        println!("  - Moyenne = {:.2}s", bounded_mean);
        println!("  - Ecart type = {:.2}s", bounded_standard_deviation);
        println!("  - Médiane = {:.2}s", bounded_median);
    }

    #[test]
    fn compute_vdf_difficulty_should_work() {
        assert_eq!(
            601,
            compute_vdf_difficulty(
                1,
                H256::from(hex!(
                    "9fa752911d55c3a1246133fe280785afbdba41f357e9cae1131d5f5b0a078b9c"
                )),
                U256::from(100_000_u128),
            )
        );
        assert_eq!(
            416,
            compute_vdf_difficulty(
                100_000,
                H256::from(hex!(
                    "977c701ceeaeeca4a3693141b0016c1e0a743f465b2d560023a1cbb1eb1f5cd6"
                )),
                U256::from(1_210_000_u128),
            )
        );
    }

    #[test]
    fn compute_vdf_difficulty_overflow() {
        assert_eq!(
            2,
            compute_vdf_difficulty(u64::MAX - 1, H256::repeat_byte(0xFF), U256::ONE)
        )
    }

    #[test]
    fn compute_vdf_difficulty_max() {
        assert_eq!(
            MAXIMUM_ITERATIONS - 1,
            compute_vdf_difficulty(
                1,
                H256::from(hex!(
                    "000000000000000000000000000000000000000000000000000000002FAF07FF"
                )),
                U256::ONE,
            )
        )
    }

    #[tokio::test]
    async fn vdf_should_work() {
        // ARRANGE
        // ACT
        let proof = vdf(b"\xaaadf".to_vec(), 63).await.unwrap();

        // ASSERT
        assert!(hex::encode(proof).len() > 0);
    }

    #[tokio::test]
    // Wesolowski works for any kind of difficulty.
    #[ignore]
    async fn vdf_should_not_work_int_size_too_high() {
        // ARRANGE
        // ACT
        let result = vdf(b"\xaaadf".to_vec(), 63).await;

        // ASSERT
        assert!(result.is_err());
    }

    // #[test]
    // fn vdf_from_delay_should_work() {
    //     // ARRANGE
    //     // ACT
    //     let proof = vdf_from_delay(b"\xaaadf", 1).unwrap();

    //     // ASSERT
    //     assert!(hex::encode(proof).len() > 0);
    // }

    // #[ignore]
    // #[test]
    // fn vdf_from_delay_should_overflow() {
    //     // ARRANGE
    //     // ACT
    //     let proof = vdf_from_delay(b"\xaaadf", u64::MAX).unwrap();

    //     // ASSERT
    //     assert!(hex::encode(proof).len() > 0);
    // }
}
